﻿// Copyright (C) 2003, 2005 Daisuke Arai <darai@users.sourceforge.jp>
// Copyright (C) 2008, 2013 panacoran <panacoran@users.sourceforge.jp>
// 
// This program is part of Protra.
//
// Protra is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
// 
// $Id: WhileStatementNode.cs 472 2013-06-26 15:26:49Z panacoran $

using System.Collections.Generic;

namespace Protra.Lib.Lang
{
    /// <summary>
    /// while文のノードを表すクラス。
    /// </summary>
    public class WhileStatementNode : ExecutableNode
    {
        /// <summary>
        /// 条件式のノード
        /// </summary>
        private ExpressionNode _condNode;

        /// <summary>
        /// ノードのリスト
        /// </summary>
        private readonly List<ExecutableNode> _nodeList = new List<ExecutableNode>();

        /// <summary>
        /// 構文解析する。
        /// </summary>
        /// <exception cref="Protra.Lib.Lang.ParseException">
        /// 構文解析中にエラーが発生したときにthrowされる。
        /// </exception>
        /// <return>
        /// 解析した結果のノード。
        /// </return>
        public override ExecutableNode Parse(Scanner scanner)
        {
            Token = scanner.Token;

            if (!scanner.Scan())
                throw new ParseException("expression is expected", Token);
            _condNode = new ExpressionNode().Parse(scanner);
            if (scanner.Token.Value != ";")
                throw new ParseException("';' expected", Token);
            while (scanner.Scan())
            {
                if (scanner.Token.Value == "end")
                {
                    scanner.Scan();
                    if (scanner.Token.Value != ";")
                        throw new ParseException("';' expected", Token);
                    return this;
                }
                _nodeList.Add(new StatementNode().Parse(scanner));
            }
            throw new ParseException("unexpected EOF", Token);
        }

        /// <summary>
        /// プログラムを実行する。
        /// </summary>
        /// <exception cref="Protra.Lib.Lang.RuntimeException">
        /// プログラム実行中にエラーが発生した場合にthrowされる。
        /// </exception>
        /// <param name="resource">リソース</param>
        /// <param name="at">int型@作用素の値</param>
        /// <param name="ats">string型@作用素の値</param>
        /// <returns>実行結果</returns>
        public override Result Execute(Resource resource, int at, string ats)
        {
            var val = _condNode.Evaluate(resource, at, ats);
            while (val != null && val.IsTrue)
            {
                foreach (var node in _nodeList)
                {
                    switch (node.Execute(resource, at, ats))
                    {
                        case Result.Return:
                            return Result.Return;
                        case Result.Break:
                            return Result.Normal;
                        case Result.Confinue:
                            goto Continue;
                    }
                }
                Continue:
                val = _condNode.Evaluate(resource, at, ats);
            }
            return Result.Normal;
        }
    }
}